#ifndef __ARITHMAT_H
#define __ARITHMAT_H

#include <iostream>
#include <string.h>
#include "ArithVec.h"

/** Basic class template for matrices.
		
		In this template a matrix is defined as an array of vectors. This may
		not in all cases be the most efficient but it has the advantage that 
		it is possible to use the double subscripting notation:
		
		T x = m[i][j]

		This template should be used through inheritance just like the 
		vector template */
template <class VVT, class HVT, class MT, int ROWS>
class ArithMat: public ArithVec<HVT,MT,ROWS> 
{ 
public:

	/// The type of a matrix element
	typedef typename HVT::ScalarType ScalarType;

	/// Get vertical dimension of matrix 
	static int get_v_dim() {return VVT::get_dim();}

	/// Get horizontal dimension of matrix
	static int get_h_dim() {return HVT::get_dim();}

	/** Set values by passing an array to the matrix.
			The values should be ordered like [[row][row]...[row]] */
	void set_array(const ScalarType* sa) 
	{
		memcpy(&(*this)[0][0], sa, 
					 sizeof(ScalarType)*get_h_dim()*get_v_dim());
	}
	
	/// Get pointer to values. See the above documentation for set_array
	const ScalarType* get_array() const 
	{
		return (*this)[0][0];
	}

	/// Default constructor. Constructs NULL matrix
  ArithMat(): 
		ArithVec<HVT,MT,ROWS>(HVT()) {}

	/// Construct Nx2 Matrix from two vectors
  ArithMat(HVT _a, HVT _b): 
		ArithVec<HVT,MT,ROWS>(_a,_b) {}

	/// Construct Nx3 Matrix from three vectors
  ArithMat(HVT _a, HVT _b, HVT _c): 
		ArithVec<HVT,MT,ROWS>(_a,_b,_c) {}

	/// Construct Nxf Matrix from four vectors
  ArithMat(HVT _a, HVT _b, HVT _c, HVT _d): 
		ArithVec<HVT,MT,ROWS>(_a,_b,_c,_d) {}

	/// Construct matrix from array of values.
	ArithMat(const ScalarType* sa): 
		ArithVec<HVT,MT,ROWS>(HVT())
	{
		set_array(sa);
	}
  
	/// Multiply two matrices.
  MT operator*(const MT& m) const;

	/// Multiply vector onto matrix
  VVT operator*(const HVT& v) const;

	/// Assigment multiplication ... probably inefficient.
  void operator *= (const MT& m)
    {
      *this = *this * m;
    }
};

template <class VVT, class HVT, class MT, int ROWS>
inline VVT ArithMat<VVT,HVT,MT,ROWS>::operator*(const HVT& v) const
{
	VVT v2;
  for(int i=0;i<ROWS;i++) v2[i] = dot((*this)[i], v);
  return v2;
}


template <class VVT, class HVT, class MT, int ROWS>
inline MT ArithMat<VVT,HVT,MT,ROWS>::operator*(const MT& m) const
{
	assert(ROWS==HVT::get_dim());
  MT n;
  for(int i=0;i<ROWS;i++)
    for(int j=0;j<ROWS;j++)
      for(int k=0;k<ROWS;k++)
        n[i][j] += (*this)[i][k] * m[k][j]; 
  
  return n;
}

template <class VVT, class HVT, class MT, int ROWS>
std::ostream& operator<<(std::ostream&os, 
												 const ArithMat<VVT,HVT,MT,ROWS>& m)
{
  os << "[\n";
  for(int i=0;i<ROWS;i++) os << "  " << m[i] << "\n";
  os << "]\n";
  return os;
}

template <class VVT, class HVT, class MT, int ROWS>
std::istream& operator>>(std::istream&is, 
												 const ArithMat<VVT,HVT,MT,ROWS>& m)
{
  for(int i=0;i<ROWS;i++) is>>m[i];
  return is;
}

template <class M1T, class M2T> 
inline M2T transpose(const M1T& m)
{
  M2T new_mat;

	assert(M1T::get_h_dim() == M2T::get_v_dim());
	assert(M2T::get_h_dim() == M1T::get_v_dim());
	
  for(int i=0;i<M2T::get_v_dim();i++)
    for(int j=0;j<M2T::get_h_dim();j++)
      new_mat[i][j] = m[j][i];
  return new_mat;
}

#endif
