#ifndef __ARITHVEC4_H
#define __ARITHVEC4_H

#include <assert.h>
#include <iostream>

template <class T, class V> class ArithVec4
{
protected:
  
  /// The coordinates of a 4D arithmetic vector.
  T a,b,c,d;

public:
  
  /// Constructing a 4D arithmetic vector from its coordinates
  ArithVec4(T _a, T _b, T _c, T _d):  a(_a),b(_b),c(_c),d(_d)
    {}

  /// The default constructor creates a NIL vector
  ArithVec4(): a(T(0)), b(T(0)), c(T(0)), d(T(0))) 
    {}

  /// Index operator used for assignment
  T& operator [] (int);

  /// Index operator
  const T& operator [] (int) const;  

  /// Equal (compares all coordinates and combines with `&&')
  bool operator == (const V& ) const;

  /// Not equal
  bool operator != (const V& ) const;
  
  /// addition
  V operator + ( const V& ) const;

  /// subtraction
  V operator - ( const V& ) const;

  /// division by vector, divides each coordinate separately
  V operator / (const V& ) const;

  /// negation
  V operator - () const;

  /// multiplication with scalar
  V operator * (const T& ) const;

  /// division by scalar
  V operator / (const T& ) const;
    
  /// Assignment subtraction
  void operator -= ( const V& );

  /// Assignment addition
  void operator += ( const V& );

  /** Assignment multiplication with vector. Multiplies each
      coordinate. */
  void operator *= ( const V& );

  /// Assignment division with vector. Multiplies each coordinate
  void operator /= ( const V&);
  
  /// Assignment multiplication with scalar.
  void operator *= ( const T& );

  /// Assignment division with scalar.
  void operator /= ( const T& );
  
  /** Less than (compares all coordinates and combines with `&&')

      This '<' function cannot be used as an ``ordering relation''
      because if a < b is false that does not imply that a>=b.
      Rather a < b means that all three cordinates of b are greater than
      those of a, and that means that b is on one side of three axis
      aligned planes that intersect at a.

      Useful since   a<b && b<c means that b is in a box where a is 
      the corner of smallest coordinates and c is the corner of largest
      coordinates. The other functions all_le, all_g, all_ge are similar */
  bool  all_l  (const V& ) const; 

  /// Less than or equal (compares all coordinates and combines with `&&')
  bool  all_le (const V& ) const;

  /// Greater than (compares all coordinates and combines with `&&')
  bool  all_g  (const V& ) const;

  /// Greater than or equal (compares all coordinates and combines with `&&')
  bool  all_ge (const V& ) const;

  /// Less than operator for use in ordering relations. 
  bool operator <(const V&) const;

};


template <class T, class V>
inline V ArithVec4<T,V>::operator + ( const V& x ) const 
{
  return V( x.a+a, x.b+b, x.c+c, x.d+d);
}

template <class T, class V>
inline V ArithVec4<T,V>::operator - ( const V& x ) const 
{
  return V( a-x.a, b-x.b, c-x.c, d-x.d );
}

template <class T, class V>
inline V ArithVec4<T,V>::operator - ( void ) const 
{
  return V( -a, -b, -c, -d ); 
}

template <class T, class V>
inline V ArithVec4<T,V>::operator * ( register const T& k ) const 
{
  return V( a*k, b*k, c*k, d*k );
}

template <class T, class V>
inline V ArithVec4<T,V>::operator / ( register const T& k ) const 
{
  return V( a/k, b/k, c/k, d/k );
}

template <class T, class V>
inline V ArithVec4<T,V>::operator / (const V& x) const
{
  return V(a/x.a, b/x.b, c/x.c, d/x.d );
}


template <class T, class V>
inline void ArithVec4<T,V>::operator -= ( const V& x ) 
{ 
  a-=x.a; 
  b-=x.b; 
  c-=x.c;
  d-=x.d;
}

template <class T, class V>
inline void ArithVec4<T,V>::operator += ( const V& x ) 
{
  a+=x.a; 
  b+=x.b; 
  c+=x.c; 
  d+=x.d; 
}

template <class T, class V>
inline void ArithVec4<T,V>::operator *= ( const V& x ) 
{ 
  a*=x.a; 
  b*=x.b; 
  c*=x.c; 
  d*=x.d; 
}

template <class T, class V>
inline void ArithVec4<T,V>::operator /= ( const V& x ) 
{ 
  a/=x.a; 
  b/=x.b; 
  c/=x.c; 
  d/=x.d; 
}


template <class T, class V>
inline void ArithVec4<T,V>::operator *= ( const T& k ) 
{ 
  a*=k; 
  b*=k; 
  c*=k; 
  d*=k; 
}

template <class T, class V>
inline void 
ArithVec4<T,V>::operator /= ( const T& k ) 
{ 
  a/=k; 
  b/=k; 
  c/=k; 
  d/=k;
}

template <class T, class V>
inline bool ArithVec4<T,V>::all_le (const V& x) const 
{
  return (a<=x.a) && (b<=x.b) && (c<=x.c) && (d<=x.d);
}

template <class T, class V>
inline bool ArithVec4<T,V>::all_l (const V& x) const 
{
  return (a<x.a && b<x.b && c<x.c && d<x.d);
}


template <class T, class V>
inline bool ArithVec4<T,V>::all_g (const V&x) const 
{
  return (a>x.a && b>x.b && c>x.c && d>x.d);
}

template <class T, class V>
inline bool ArithVec4<T,V>::all_ge ( const V& x ) const 
{
  return (a>=x.a && b>=x.b && c>=x.c && d>=x.d);
}

template <class T, class V>
inline bool ArithVec4<T,V>::operator <(const V& x) const
{
  return a==x.a?(b==x.b?(c==x.c?(d==x.d?false:d<x.d):c<x.c):b<x.b):a<x.a;
}

template <class T, class V>
inline bool ArithVec4<T,V>::operator ==(const V& x) const
{
  return a==x.a && b==x.b && c==x.c && d==x.d;
}


template <class T, class V>
inline T& ArithVec4<T,V>::operator [] ( int index )
{
  assert(index<4);
  switch( index ) 
    {
    case 0: return a;
    case 1: return b;
    case 2: return c;
    case 3: return d;
    }
  return d;
}

template <class T, class V>
inline const T& 
ArithVec4<T,V>::operator [] ( int index ) const
{
  assert(index<4);
  switch( index ) 
    {
    case 0: return a;
    case 1: return b;
    case 2: return c;
    case 3: return d;
    }
  return d;
}

template <class T,class V> 
inline std::ostream& operator<<( std::ostream& os, const ArithVec4<T,V>& x )
{
  os << " [ " 
     << x[0] << " " << x[1] << " " << x[2] << " " << x[3]
     << " ] ";
  return os;
}

template <class T, class V> 
inline std::istream& operator>>( std::istream& is, ArithVec4<T,V>& x )
{
  is >> x[0] >> x[1] >> x[2] >> x[3];
  return is;
}


#endif
