#include "Mat4x4f.h"

using namespace CGLA;

Mat4x4f invert(const Mat4x4f& this_mat) 
{
  //   The following code has been copied from a gem in Graphics Gems II by
  //   Kevin Wu.                      
  //   The function is very fast, but it can only invert affine matrices. An
  //   exception NotAffine is thrown if the matrix is not affine, and another
  //   exception Singular is thrown if the matrix is singular.

  Mat4x4f new_mat;
  float    det_1;
  float    pos, neg, temp;
      

  if (!((this_mat[0][3] == this_mat[1][3] == this_mat[2][3] == 0) && 
	(this_mat[3][3] ==1.0)))
    throw(Mat4x4fNotAffine("Can only invert affine matrices"));
    
#define ACCUMULATE if (temp >= 0.0) pos += temp; else neg += temp
  
  /*
   * Calculate the determinant of submatrix A and determine if the
   * the matrix is singular as limited by the float precision
   * floating-point this_mat representation.
   */
  
  pos = neg = 0.0;
  temp =  this_mat[0][0] * this_mat[1][1] * this_mat[2][2];
  ACCUMULATE;
  temp =  this_mat[0][1] * this_mat[1][2] * this_mat[2][0];
  ACCUMULATE;
  temp =  this_mat[0][2] * this_mat[1][0] * this_mat[2][1];
  ACCUMULATE;
  temp = -this_mat[0][2] * this_mat[1][1] * this_mat[2][0];
  ACCUMULATE;
  temp = -this_mat[0][1] * this_mat[1][0] * this_mat[2][2];
  ACCUMULATE;
  temp = -this_mat[0][0] * this_mat[1][2] * this_mat[2][1];
  ACCUMULATE;
  det_1 = pos + neg;
  
  /* Is the submatrix A singular? */
  if ((det_1 == 0.0) || (fabs(det_1 / (pos - neg)) < MINUTE)) 
    {
      /* Mat4x4f M has no inverse */
      throw(Mat4x4fSingular("Tried to invert Singular matrix"));
    }

  else {

    /* Calculate inverse(A) = adj(A) / det(A) */
    det_1 = 1.0 / det_1;
    new_mat[0][0] =   ( this_mat[1][1] * this_mat[2][2] -
                             this_mat[1][2] * this_mat[2][1] )
      * det_1;
    new_mat[1][0] = - ( this_mat[1][0] * this_mat[2][2] -
                             this_mat[1][2] * this_mat[2][0] )
      * det_1;
    new_mat[2][0] =   ( this_mat[1][0] * this_mat[2][1] -
                             this_mat[1][1] * this_mat[2][0] )
      * det_1;
    new_mat[0][1] = - ( this_mat[0][1] * this_mat[2][2] -
                             this_mat[0][2] * this_mat[2][1] )
      * det_1;
    new_mat[1][1] =   ( this_mat[0][0] * this_mat[2][2] -
                             this_mat[0][2] * this_mat[2][0] )
      * det_1;
    new_mat[2][1] = - ( this_mat[0][0] * this_mat[2][1] -
                             this_mat[0][1] * this_mat[2][0] )
      * det_1;
    new_mat[0][2] =   ( this_mat[0][1] * this_mat[1][2] -
                             this_mat[0][2] * this_mat[1][1] )
      * det_1;
    new_mat[1][2] = - ( this_mat[0][0] * this_mat[1][2] -
                             this_mat[0][2] * this_mat[1][0] )
      * det_1;
    new_mat[2][2] =   ( this_mat[0][0] * this_mat[1][1] -
                             this_mat[0][1] * this_mat[1][0] )
      * det_1;

    /* Calculate -C * inverse(A) */
    new_mat[3][0] = - ( this_mat[3][0] * new_mat[0][0] +
                             this_mat[3][1] * new_mat[1][0] +
                             this_mat[3][2] * new_mat[2][0] );
    new_mat[3][1] = - ( this_mat[3][0] * new_mat[0][1] +
                             this_mat[3][1] * new_mat[1][1] +
                             this_mat[3][2] * new_mat[2][1] );
    new_mat[3][2] = - ( this_mat[3][0] * new_mat[0][2] +
                             this_mat[3][1] * new_mat[1][2] +
                             this_mat[3][2] * new_mat[2][2] );
    
    /* Fill in last column */
    new_mat[0][3] = new_mat[1][3] = new_mat[2][3] = 0.0;
    new_mat[3][3] = 1.0;

    return new_mat;
	
  }

#undef ACCUMULATE
}




Mat4x4f rotation_matrix(Axis axis, float angle)
{
  Mat4x4f m;

  switch(axis)
    {
    case XAXIS:
      m[0][0] = 1.0;
      m[1][1] = cos(angle);
      m[1][2] = sin(angle);
      m[2][1] = -sin(angle);
      m[2][2] = cos(angle);
      m[3][3] = 1.0;
      break;
    case YAXIS:
      m[0][0] = cos(angle);
      m[0][2] = -sin(angle);
      m[2][0] = sin(angle);
      m[2][2] = cos(angle);
      m[1][1] = 1.0;
      m[3][3] = 1.0;
      break;
    case ZAXIS:
      m[0][0] = cos(angle);
      m[0][1] = sin(angle);
      m[1][0] = -sin(angle);
      m[1][1] = cos(angle);
      m[2][2] = 1.0;
      m[3][3] = 1.0;
      break;
    }

  return m;
}

Mat4x4f translation_matrix(const Vec3f& v)
{
  Mat4x4f m;

  m[0][0] = 1.0;
  m[1][1] = 1.0;
  m[2][2] = 1.0;
  m[3][3] = 1.0;
  
  m[0][3] = v[0];
  m[1][3] = v[1];
  m[2][3] = v[2];
  
  return m;
}

Mat4x4f scaling_matrix(const Vec3f& v)
{
  Mat4x4f m;

  m[0][0] = v[0];
  m[1][1] = v[1];
  m[2][2] = v[2];
  m[3][3] = 1.0;
   
  return m;
}



Mat4x4f perspective_matrix(float d)
{
  Mat4x4f m;
  
  /* Eye at the origin, looking down the negative z axis */

  m[0][0] = 1.0;
  m[1][1] = 1.0;
  m[2][2] = 1.0;
  m[3][2] = -1.0/d;
   
  return m;
}


Mat4x4f identity_matrix()
{
  Mat4x4f m;

  m[0][0] = 1.0;
  m[1][1] = 1.0;
  m[2][2] = 1.0;
  m[3][3] = 1.0;
    
  return m;
}




