#ifndef __STLFIX_H
#define __STLFIX_H

#ifndef WIN32

/* ----------------------------------------------------------------------

	 The code below contains fixes to certain problems in the STL. The 
	 main parts of the code is adapted from the examples of Volker Simonis
	 associated with the paper:

	 Adapters and Binders - Overcoming problems 
	 in the design and implementation of the C++-STL

	 Include this file after all STL includes - or at least after <functional>
	 This is important because this file contains specializations of templates 
	 found in <functional> and the specialization must come after the 
	 definition.

	 I have also added some function and class templates that I have 
	 found useful in association with the STL.

	 ----------------------------------------------------------------------
	 The issues addressed by Simonis:
	 
	 Basically there are two issues: If a function object is created from a
	 member function (using mem_fun or mem_fun_ref) that takes a reference
	 argument, this reference argument cannot be bound, using bind2nd. The 
	 solution lies in a specialization of the binary_function template.
	 
	 The other issue is similar. The binder2nd is used to bind the second
	 argument of a functional object taking two args. Unfortunately,
	 binder2nd forces the first argument to be const, and this is not
	 always the case. The solution is a specialization of binder2nd wrt
	 mem_fun1_ref_t.

	 ----------------------------------------------------------------------
	 My own STL additions:

	 Some of my own additions are also in this file. Of course, they are non
	 standard albeit (so far) in namespace std:

	 compose_binary(f,g,h): x,y -> f(g(x), h(y))
  	compose_unary(f,g,h):   x -> f(g(x), h(x))
    compose_unary(f,g):     x -> f(g(x))

	 ptr_to_ref:          T* x  -> T& x

	 Andreas, May 2000

---------------------------------------------------------------------- */

#if ((defined __GNUC__) && (__GNUC_MINOR__ > 91)) || (!defined __GNUC__)
namespace std {
#endif

	template <class Arg1, class Arg2, class Result>
	struct binary_function<Arg1&, Arg2, Result> {
    typedef Arg1 first_argument_type;
    typedef Arg2 second_argument_type;
    typedef Result result_type;
	};      

	template <class Arg1, class Arg2, class Result>
	struct binary_function<Arg1, Arg2&, Result> {
    typedef Arg1 first_argument_type;
    typedef Arg2 second_argument_type;
    typedef Result result_type;
	};      

	template <class Arg1, class Arg2, class Result>
	struct binary_function<Arg1&, Arg2&, Result> {
    typedef Arg1 first_argument_type;
    typedef Arg2 second_argument_type;
    typedef Result result_type;
	};      

	// This is in fact a problem of older STL-versions shipped with g++
#if (defined __GNUC__) && (__GNUC_MINOR__ < 95) 
	template <class S, class T, class A>
	inline mem_fun1_ref_t<S,T,A> mem_fun_ref(S (T::*f)(A)) {
		return mem_fun1_ref_t<S,T,A>(f);
	}
#endif

	template <class A, class B, class C> 
	class binder2nd<mem_fun1_ref_t<A,B,C> > :
		public unary_function<typename mem_fun1_ref_t<A,B,C>::first_argument_type,
													typename mem_fun1_ref_t<A,B,C>::result_type> 
	{
	protected:
		mem_fun1_ref_t<A,B,C> op;
		typename mem_fun1_ref_t<A,B,C>::second_argument_type value;
	public:
		binder2nd(const mem_fun1_ref_t<A,B,C>& x,
							const typename mem_fun1_ref_t<A,B,C>::second_argument_type& y) 
      : op(x), value(y) {}

		typename mem_fun1_ref_t<A,B,C>::result_type
		operator()(typename mem_fun1_ref_t<A,B,C>::first_argument_type& x) const
		{
			return op(x, value); 
		}
	};

	template <class A, class B, class C> 
	class binder1st<mem_fun1_t<A,B,C> >: public 
	unary_function<typename mem_fun1_t<A,B,C>::second_argument_type,
								 typename mem_fun1_t<A,B,C>::result_type> 
	{
		typedef mem_fun1_t<A,B,C> _Operation;
	protected:
		_Operation op;
		typename _Operation::first_argument_type value;
	public:
		binder1st(const _Operation& __x,
							const typename _Operation::first_argument_type& __y)
      : op(__x), value(__y) {}
		typename _Operation::result_type
		operator()(typename _Operation::second_argument_type& __x) const {
			return op(value, __x); 
		}
	};


	// The one below is my own - it converts pointers to references. It should
	// not be necessary to create such a function oneself, so I call it a fix.

	template<class T>
	struct ptr_to_ref
	{
		T& operator()(T* s) {return (*s);}
	};


	template<class BFUN, class UFUN1, class UFUN2>
	class BinFunComposition_bin_of_2una:
		public binary_function<typename UFUN1::argument_type,
													 typename UFUN2::argument_type,
													 typename BFUN::result_type>
	{
		typedef typename UFUN1::argument_type ARG1;
		typedef typename UFUN2::argument_type ARG2;
		typedef typename BFUN::result_type    RES;
	
		BFUN bin_fun;
		UFUN1 una_fun1;
		UFUN2 una_fun2;
	public:
	
		BinFunComposition_bin_of_2una(const BFUN& _bin_fun, 
																	const UFUN1& _una_fun1, 
																	const UFUN2& _una_fun2): 
			bin_fun(_bin_fun), una_fun1(_una_fun1), una_fun2(_una_fun2) {}

		RES operator()(const ARG1& arg1, const ARG2& arg2) const 
		{
			return (bin_fun)( una_fun1(arg1), una_fun2(arg2) );
		}
	};

	template<class BFUN, class UFUN1, class UFUN2>
	class UnaFunComposition_bin_of_2una:
		public unary_function<typename UFUN1::argument_type,
													typename BFUN::result_type>
	{
		typedef typename UFUN1::argument_type ARG;
		typedef typename BFUN::result_type    RES;
	
		BFUN bin_fun;
		UFUN1 una_fun1;
		UFUN2 una_fun2;
	public:
	
		UnaFunComposition_bin_of_2una(const BFUN& _bin_fun, 
																	const UFUN1& _una_fun1, 
																	const UFUN2& _una_fun2): 
			bin_fun(_bin_fun), una_fun1(_una_fun1), una_fun2(_una_fun2) {}

		RES operator()(const ARG& arg) const 
		{
			return (bin_fun)( una_fun1(arg), una_fun2(arg) );
		}
	};


	template<class UFUN1, class UFUN2>
	class UnaFunComposition_una_of_una:
		public unary_function<typename UFUN2::argument_type,
   												typename UFUN1::result_type>
	{
		typedef typename UFUN2::argument_type ARG;
		typedef typename UFUN1::result_type   RES;
	
		UFUN1 una_fun1;
		UFUN2 una_fun2;
	public:
	
		UnaFunComposition_una_of_una(const UFUN1& _una_fun1, 
																 const UFUN2& _una_fun2): 
			una_fun1(_una_fun1), 
			una_fun2(_una_fun2) {}

		RES operator()(ARG& arg) const
		{
			return una_fun1(una_fun2(arg));
		}
	};

	template<class UFUN1, class FUN2_ARG, class FUN2_RES>
	class UnaFunComposition_una_of_una<UFUN1,const_mem_fun_t<FUN2_ARG,FUN2_RES> >
		:public unary_function<
      	typename const_mem_fun_t<FUN2_ARG,FUN2_RES>::argument_type,
				typename UFUN1::result_type>
	{
		typedef const_mem_fun_t<FUN2_ARG,FUN2_RES> UFUN2;
		typedef typename UFUN2::argument_type ARG;
		typedef typename UFUN1::result_type   RES;
	
		UFUN1 una_fun1;
		UFUN2 una_fun2;
	public:
	
		UnaFunComposition_una_of_una(const UFUN1& _una_fun1, 
																 const UFUN2& _una_fun2): 
			una_fun1(_una_fun1), 
			una_fun2(_una_fun2) {}

		RES operator()(const ARG& arg) const
		{
			return una_fun1(una_fun2(arg));
		}
	};

	template<class FUN1_ARG, class FUN1_RES, class UFUN2>
	class UnaFunComposition_una_of_una<const_mem_fun_t<FUN1_ARG,FUN1_RES>,UFUN2>
		:public unary_function<
       	typename UFUN2::argument_type,
      	typename const_mem_fun_t<FUN1_ARG,FUN1_RES>::result_type>
	{
		typedef const_mem_fun_t<FUN1_ARG,FUN1_RES> UFUN1;
		typedef typename UFUN2::argument_type ARG;
		typedef typename UFUN1::result_type   RES;
	
		UFUN1 una_fun1;
		UFUN2 una_fun2;
	public:
	
		UnaFunComposition_una_of_una(const UFUN1& _una_fun1, 
																 const UFUN2& _una_fun2): 
			una_fun1(_una_fun1), 
			una_fun2(_una_fun2) {}

		RES operator()(const ARG& arg) const
		{
			return una_fun1(una_fun2(arg));
		}
	};

	template<class BFUN, class UFUN1, class UFUN2>
	inline BinFunComposition_bin_of_2una<BFUN, UFUN1, UFUN2>
	compose_binary(const BFUN& a, const UFUN1& b, const UFUN2& c) 
	{
		return BinFunComposition_bin_of_2una<BFUN, UFUN1, UFUN2>(a,b,c);
	}

	template<class BFUN, class UFUN1, class UFUN2>
	inline UnaFunComposition_bin_of_2una<BFUN, UFUN1, UFUN2>
	compose_unary(const BFUN& a, const UFUN1& b, const UFUN2& c) 
	{
		return UnaFunComposition_bin_of_2una<BFUN, UFUN1, UFUN2>(a,b,c);
	}

	template<class UFUN1, class UFUN2>
	inline UnaFunComposition_una_of_una<UFUN1, UFUN2>
	compose_unary(const UFUN1& a, const UFUN2& b) 
	{
		return UnaFunComposition_una_of_una<UFUN1, UFUN2>(a,b);
	}


	
#if ((defined __GNUC__) && (__GNUC_MINOR__ > 91)) || (!defined __GNUC__)
} // namespace std
#endif

#endif


#endif
