#ifndef __STATEKEEPER_H
#define __STATEKEEPER_H
// Author: J. Andreas Brentzen,
// Created: Thu Mar 30 12:49:0

#include "Action.h"
#include "ActionHistory.h"

/** StateKeeper is an abstract class for representing a stateful system with 
		a command history and a state stack which combine to allow undo redo
		functionaliy.
		
		Classes derived from this one should override the virtual methods for 
		pushing and popping the state of the system that the derived classes
		represent.  */
class StateKeeper
{
protected:

	ActionHistory hist;


private:
	/** Save state  */
	virtual void push_state() = 0;

	/** Restore all state from stack */
	virtual bool pop_state() = 0;

	/** Some variables may need updating if we pop a previous state. 
			Even if all state should be pushed and popped there may be some 
			variables that do contain "state" but "state" which is redundant
			and can be computed from the popped state. This function recomputes
			such variables */
	virtual void reset_from_state() = 0;

	/** Compuations that are always carried out after each action can be 
			put here. For instance in a constraint based system, adding a 
			constraint should always entail a resolution of the constraint 
			system */
	virtual bool post_action() = 0;

public:	

	/// Call to undo last action. No effect, if we are at first action
	void undo_last_action();

	/// Redo last undone action. If nothing is undone it has no effect.
	void redo_last_undone_action();
	
	/** Try an action. It should leave the system in a sane state returning
			false if the action can't be carried out. */
	bool try_action(Action*);

};

#endif
