#include <GL/glut.h>
#include "TrackBall.h"

TrackBall::TrackBall(const Vec3f& _centre, float _eye_dist,
										 unsigned _width, unsigned _height):
	centre(_centre), eye_dist(_eye_dist), width(_width), height(_height)
{
  // This size should really be based on the distance from the center of
  // rotation to the point on the object underneath the mouse.  That
  // point would then track the mouse as closely as possible.  This is a
  // simple example, though, so that is left as an exercise.
  ballsize = 0.8;
  reset();
}

void TrackBall::grab_ball(TrackBallAction act, int xp, int yp)
{
	float x = scalePoint(xp, 0, width);
	float y = scalePoint(yp, height, -(long)height);
	set_position(x, y);
	current_action = act;
}

void TrackBall::roll_ball(int xp, int yp)
{
	float x = scalePoint(xp, 0, width);
	float y = scalePoint(yp, height, -(long)height);
			
	switch (current_action) 
		{
		case ROTATE_ACTION:
			rotate(x, y);
			break;
			
		case PAN_ACTION:
			pan(x, y);
			break;
			
		case ZOOM_ACTION:
			zoom(x, y);
			break;
		}
}


// Bring trackball to home position, no translation, zero rotation.
void TrackBall::reset()
{
  qrot = Quaternion(0.0, 0.0, 0.0, 1.0);
  qinc = Quaternion(0.0, 0.0, 0.0, 1.0);
  trans = Vec3f(0.0, 0.0, - eye_dist);
  scale = Vec3f(eye_dist, eye_dist, eye_dist);
}

// Modify the current gl matrix by the trackball rotation and translation.
void TrackBall::set_gl_modelview() const
{
	Mat4x4f m;
  glTranslatef(trans[0], trans[1], trans[2]);
	glTranslatef(centre[0], centre[1], centre[2]);
  m = qrot.get_mat();
  glMultMatrixf(&(m[0][0]));
	glTranslatef(-centre[0], -centre[1], -centre[2]);
	
}

// Call this when the user does a mouse down.  
// Stop the trackball glide, then remember the mouse
// down point (for a future rotate, pan or zoom).
void TrackBall::set_position(float newx, float newy) 
{
  stop_spin();
  lastx = newx;
  lasty = newy;
}

// Rotationaly spin the trackball by the current increment.
// Use this to implement rotational glide.
void TrackBall::do_spin() 
{
  qrot = qrot*qinc;
}

// Cease any rotational glide by zeroing the increment.
void TrackBall::stop_spin() 
{
  qinc.set(0.0, 0.0, 0.0, 1.0);
}

void TrackBall::rotate(float newx, float newy) 
{
  calcRotation(newx, newy, lastx, lasty);
  do_spin();

  lastx = newx;	/* remember for next time */
  lasty = newy;
}

void TrackBall::pan(float newx, float newy) 
{
  trans[0] += (newx - lastx) * scale[0];
  trans[1] += (newy - lasty) * scale[1];

  lastx = newx;	/* remember for next time */
  lasty = newy;
}

void TrackBall::zoom(float newx, float newy) 
{
  trans[2] += (newy - lasty) * scale[2];

  lastx = newx;	/* remember for next time */
  lasty = newy;
}

void TrackBall::calcRotation(float newX, float newY, float oldX, float oldY) 
{
  Vec3f	p1, p2;	/* 3D mouse points  */
  float	L;		/* sin^2(2 * phi)   */
  Vec3f q;

  // Check for zero rotation
  if (newX == oldX  &&  newY == oldY) {
    qinc.set(0.0, 0.0, 0.0, 1.0);
    return;
  }
    
  // Form two vectors based on input points, find rotation axis
  p1 = Vec3f(newX, newY, projectToSphere(newX, newY));
  p2 = Vec3f(oldX, oldY, projectToSphere(oldX, oldY));
  
  q = cross(p1, p2);		/* axis of rotation from p1 and p2 */
  
  L = dot(q,q) / (dot(p1,p1) * dot(p2,p2));
  L = sqrtf(1 - L);
  
  q.normalize();				/* q' = axis of rotation */
  q *= sqrtf((1 - L)/2);	/* q' = q' * sin(phi) */

	qinc.set(q[0],q[1],q[2],sqrtf((1 + L)/2));
}

// Project an x,y pair onto a sphere of radius r OR a hyperbolic sheet
// if we are away from the center of the sphere.
float TrackBall::projectToSphere(float x, float y) 
{
  float d, t, z;

  d = sqrt(x*x + y*y);
  
  // Inside sphere 
  if (d < ballsize * 0.70710678118654752440) {   
    z = sqrt(ballsize*ballsize - d*d);
  }
  // On hyperbola 
  else {           
    t = ballsize / 1.41421356237309504880;
    z = t*t / d;
  }

  return z;
}

// Scales integer point to the range [-1, 1]
float TrackBall::scalePoint(long pt, long origin, long size)
{
  float	x;

  x = (float) (pt - origin) / (float) size;
  if (x < 0.0) x = 0.0;
  if (x > 1.0) x = 1.0;

  return (2.0 * x) - 1.0;
}
