#ifndef __TRILINEARCELL_H
#define __TRILINEARCELL_H

#include <assert.h>
#include "Common/CommonDefs.h"
#include "CGLA/Vec3f.h"
#include "Ray.h"

/** We use the following constant as the value deemed close enough
    to zero in the TrilinearCell and derived classes. */
const float CLOSE_TO_ZERO = 5e-8;

/** This class facilitates trilinear interpolation and isosurface 
    intersection. The intersection function currently handles only
    the case where the isovalue is 0 but it is trivial to change */
class TrilinearCell
{
protected:
  
  /// Diagonally opposite corners of the cell.
  Vec3f corner0, corner7;
  
  /// The point where we interpolate a value.
  Vec3f point;

  /// Edge lengths along x,y and z axes.
  Vec3f dist;

  /// Interpolation weights.
  float alpha, beta, gamm;
  float m_alpha, m_beta, m_gamm;

#ifndef NDEBUG
  /** We might need to keep track of whether we have 
      a valid point */
  bool valid_point;
#endif
  
public:

  TrilinearCell(const Vec3f& _corner0, const Vec3f& _corner7): 
    corner0(_corner0), corner7(_corner7)
#ifndef NDEBUG
    ,valid_point(false) 
#endif
    {
      dist = corner7-corner0;
    }

  TrilinearCell(): 
    corner0(Vec3f(0,0,0)), corner7(Vec3f(1,1,1))
#ifndef NDEBUG
    ,valid_point(false) 
#endif
    {
      dist = corner7-corner0;
    }

  /** Set the point, whose value we wish to interpolate.
      Makes sure that the point is valid */
  bool safe_set_point(const Vec3f& point);

  /// Set the point, whose value we wish to interpolate.
  void set_point(const Vec3f& point);

  /// Retrieve the point whose value we interpolate.
  const Vec3f& get_point() const {return point;}

  /** Passed an array of eight values, this function returns 
      the interpolated value. Since it is a template function,
      we can interpolate everything that implements `*' between
      itself and a float */
  template<class T>
  T interpolate(T vals[8]) const
    {
      assert(valid_point);
      return 
        vals[0] * (m_alpha * m_beta * m_gamm) + 
        vals[1] * (alpha   * m_beta * m_gamm) +
        vals[2] * (m_alpha * beta   * m_gamm) + 
        vals[3] * (alpha   * beta   * m_gamm) + 
        vals[4] * (m_alpha * m_beta *   gamm) +
        vals[5] * (alpha   * m_beta *   gamm) +
        vals[6] * (m_alpha * beta   *   gamm) +
        vals[7] * (alpha   * beta   *   gamm);
    }

  
  /** Find the intersection between a ray and a trilinear cell. Passed
      eight floating point values, a point, a direction, and finally the 
      isovalue, the function returns the point of intersection.
      
      `point' and the interpolation weights are automatically set to
      the values of the found point. */
  bool intersect(float[8], Ray&);
 
};

#endif
