/**
	 Thu Nov  2 17:50:14 CET 2000

	 Author: Andreas Brentzen.
	 
	 This program is a test and demo of how to render points quickly
	 using VERTEX arrays that are stored in AGP memory. This requires
	 an NVIDIA graphics card and corresponding OpenGL libraries.
	 
	 If the prerequisites are in place, the technique allows geometry
	 to be transferred quickly from agp memory to the graphics card  
	 using DMA. The example uses glXAllocateMemoryNV rather than the 
	 windows equivalent since it is intended for the linux platform.

	 Even if it is possible to get enough memory, the employed NVIDIA
	 specific extensions have problems with vertex arrays if the number
	 of indices (i.e. in this case - simply - the number of points we
	 wanna render) is greater than 65536. Hence we divide the buffer 
	 into chunks if there are more points than 65536.

	 A technique that is not demonstrated is the use of fences when there is 
	 more geometry than the allocateable agp memory will hold. This
	 technique is in the NVIDIA example learning_VAR (available at
	 NVIDIA.com) but that example is for the windows API wgl rather than
	 glx - so this program is useful for Linux users. Anyway - I can
	 allocate a million points which is usually enough and pulls the
	 framerate down to about 10 fps.

	 To understand what is going on look at the www.NVIDIA.com developer
	 relations homepage. There are several docs about these techniques.
	 `OpenGL performance FAQ for NVIDIA GPUs v 2.0' is a good starting
	 point.

	 My understanding of the low level/hw issues is not too strong, but
	 the program does show a 2 times speedup in some cases, so you don't
	 need a perfect understanding to make it work.
*/

// For some reason (on my system) the constant below must be defined.
#define GL_GLEXT_PROTOTYPES
#include "CGLA/Vec3i.h"
#include "CGLA/Vec3f.h"
#include "Graphics/TrackBall.h"
#include <GL/gl.h>
#include <GL/glx.h>
#include <GL/glext.h>
#include <GL/glut.h>

// Some constants --------------------------------------------------

const int VECSIZE = 4;                            // Size of a vector in
                                                  // floating point units
const int XPOINTS=1000;                            // POints to draw along X
const int YPOINTS=1000;                            // Points to draw along Y
const int POINTS = XPOINTS*YPOINTS;               // Total number of points
const int CHUNKS = POINTS/65536;

//----------------------------------------------------------------------
// Note that it the total number of points exceeds 65536 the excess
// points will be silently ignored. This is semi-documented in an 
// NVIDIA paper.
//----------------------------------------------------------------------

const int WINX = 500; // X size of window
const int WINY = 500; // Y size of window
const int PACKET_SIZE  = VECSIZE*sizeof(GLfloat); //Size in bytes of a
                                                  // vertex/normal
const int NEEDED_SPACE = 2*POINTS*PACKET_SIZE;    // Total required storage
const int CHUNK_SIZE = 65536*PACKET_SIZE;

// Global variables --------------------------------------------------
void* nv_storage = 0;
int frames = 0;
int before_first_frame=0;
TrackBall ball(Vec3f(WINX/2,WINX/2,0), WINX, WINX, WINY);


// GLUT Callbacks --------------------------------------------------
void key(unsigned char,int,int)
{
	int after_last_frame = glutGet(GLUT_ELAPSED_TIME);
	float dtime = (after_last_frame-before_first_frame)/1000;
	cout << "fps=" << frames/dtime << endl;
	cout << "point/sec=" << (frames/dtime)*POINTS << endl;
	exit(0);
}

void mouse_button(int button, int state, int x, int y)
{
	switch(state)
		{
 		case GLUT_DOWN:
			ball.grab_ball(ROTATE_ACTION,x,y);
			break;
 		case GLUT_UP:
			ball.release_ball();
			break;
 		}
	glutPostRedisplay();
}

void movefun(int x, int y)
{
	ball.roll_ball(x,y);
	glutPostRedisplay();
}

inline Vec3f* v(char* b,int i) 
{
	return (Vec3f*)(b+PACKET_SIZE*i);
}

void display()
{
 	static char* gl_vertices= (char*)(nv_storage);
 	static char* gl_normals = (char*)(nv_storage) + POINTS*PACKET_SIZE;
	
	if(frames++==0) 
		{
			before_first_frame = glutGet(GLUT_ELAPSED_TIME);
			int k=0;
			int chunk=0;
			for(int j=0;j<YPOINTS;j++)
				for(int i=0;i<XPOINTS;i++)
					{
						const float k0 = float(WINY/4);
						const float k1 = 2;
						const float k2 = 14;
						float s = float(i)/float(XPOINTS);
						float t = float(j)/float(YPOINTS);
						Vec3f p( WINX* s,WINY* t,k0*cos(k1*s)*sin(k2*t));
						Vec3f ps(WINX,0,k0*-sin(k1*s)*k1*sin(k2*t));
						Vec3f pt(0,WINY,k0*cos(k1*s)*cos(k2*t)*k2);
						Vec3f n = cross(ps,pt);
						n.normalize();
						*v(gl_vertices,k) = p;
						*v(gl_normals,k) = n;
						k++;
					}
		}
	glClear(GL_COLOR_BUFFER_BIT);
	glPushMatrix();
	ball.set_gl_modelview();

#ifdef USE_NV	
	/* If we use the AGP storage, we cannot have indices greater than 
		 65536 so we sequentially draw chunks of 65536 points */
	for(int i=0;i<=CHUNKS;i++)
		{
			glVertexPointer(3,GL_FLOAT,PACKET_SIZE, gl_vertices+i*CHUNK_SIZE);
			glNormalPointer(  GL_FLOAT,PACKET_SIZE, gl_normals+i*CHUNK_SIZE);
			if(i==CHUNKS)
				glDrawArrays(GL_POINTS, 0, POINTS%65536);
			else
				glDrawArrays(GL_POINTS, 0, 65536);
		}
#else
			glVertexPointer(3,GL_FLOAT,PACKET_SIZE, gl_vertices);
			glNormalPointer(  GL_FLOAT,PACKET_SIZE, gl_normals);
			glDrawArrays(GL_POINTS, 0, POINTS);
#endif
	glPopMatrix();
	glutSwapBuffers();

}


void idle()
{
	ball.do_spin();
	glutPostRedisplay();
}


// MAIN --------------------------------------------------
int main(int argc, char *argv[])
{
  glutInit( &argc, argv );
	glutInitDisplayMode(GLUT_RGBA|GLUT_DOUBLE);
  glutInitWindowSize( WINX, WINY);
  glutCreateWindow(argv[0]);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,WINX,0,WINY,-2*WINX,2*WINX); 
  glMatrixMode(GL_MODELVIEW);
	
	glutMotionFunc(movefun);
	glutMouseFunc(mouse_button);
  glutDisplayFunc(display);
  glutIdleFunc(idle);
	glutKeyboardFunc(key);

#ifdef USE_NV
 	typedef void *(*glxamType)(int,float,float,float);
 	glxamType glxam_nv;
	glxam_nv = (glxamType)glXGetProcAddressARB((const GLubyte*)"glXAllocateMemoryNV");
 	nv_storage = (*glxam_nv)(NEEDED_SPACE,0.25,0.25,.75);
	if(nv_storage==0)
		{
			cout << "arrgh didn't get storage " << endl;
			exit(0);
		}
	cout << "Got " << NEEDED_SPACE << " bytes " << endl;
	glVertexArrayRangeNV(NEEDED_SPACE, nv_storage);	
 	glEnableClientState(GL_VERTEX_ARRAY_RANGE_NV);
#else
	static char block[NEEDED_SPACE];
	nv_storage = (void*) block;
#endif
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glClearColor(0.3,0.3,0.3,0);
	glPointSize(3);
	GLfloat matAmb[4]  = { 0.1, 0.2 , 0.4, .5};
	GLfloat matDiff[4] = { 0.2, 0.4 , 0.8, .5};
  GLfloat matSpec[4] = { .1,.3,.8, .5};
  GLfloat matShine = 10.00;
  
  glMaterialfv(GL_FRONT, GL_DIFFUSE, matAmb);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, matDiff);
  glMaterialfv(GL_FRONT, GL_SPECULAR, matSpec);
  glMaterialf(GL_FRONT, GL_SHININESS, matShine);

  glutMainLoop();
  return 1;
}
